/*

    File: analyse.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */
#include <stdio.h>
#include <string.h>
#ifndef  DJGPP
#include <sys/types.h>
#endif
#include "types.h"
#include "common.h"
#include "fnctdsk.h" /* get_LBA_part */
#include "hdaccess.h"
#include "analyse.h"
#include "bfs.h"
#include "swap.h"
#include "bsd.h"
#include "fat.h"
#include "ntfs.h"
#include "ext2.h"
#include "netware.h"
#include "intrface.h"
#include "rfs.h"
#include "lvm.h"

int search_type_0(t_param_disk disk_car,t_diskext partition,const int debug, const int dump_ind,const int paranoid)
{
  uchar buffer[8*0x200];
  loff_t hd_offset=partition->lba;
  if(hd_read_LBA(disk_car,8, &buffer, hd_offset))
    return -1;
  if((LBA2sector(disk_car,partition->lba)==disk_car->CHS.sector) && (LBA2head(disk_car,partition->lba)==disk_car->CHS.head))
  {	/* NTFS recovery using backup sector */
    if(!recover_NTFS(disk_car,(struct ntfs_boot_sector*)&buffer,partition,debug,dump_ind))
    {
      partition->lba=partition->lba-partition->part_size+1;
      strcpy(partition->info,"NTFS found using backup sector!");
//      ecrit_rapport("analyse.c : NTFS found using backup sector!");
      return 1;
    }
    return 0;
  }
  if(LBA2sector(disk_car,partition->lba)==(1+6))
  {	/* FAT32 recovery using backup sector */
    if(!recover_FAT32(disk_car,(struct fat_boot_sector*)&buffer,partition,debug,dump_ind))
    {
      partition->lba-=6;	/* backup sector ... */
      strcpy(partition->info,"FAT found using backup sector!");
      return 1;
    }
    return 0;
  }
  if(!recover_Linux_SWAP(disk_car,(union swap_header *)&buffer,partition,debug,dump_ind)) return 1;
  if(!recover_LVM(disk_car,(pv_disk_t*)&buffer,partition,debug,dump_ind)) return 1;
  if(!recover_FAT12(disk_car,(struct fat_boot_sector*)&buffer,partition,debug,dump_ind)) return 1;
  if(!recover_FAT16(disk_car,(struct fat_boot_sector*)&buffer,partition,debug,dump_ind)) return 1;
  if(!recover_FAT32(disk_car,(struct fat_boot_sector*)&buffer,partition,debug,dump_ind)) return 1;
  if(!recover_HPFS(disk_car,(struct fat_boot_sector*)&buffer,partition,debug,dump_ind)) return 1;
  if(!recover_OS2MB(disk_car,(struct fat_boot_sector*)&buffer,partition,debug,dump_ind)) return 1;
  if(!recover_NTFS(disk_car,(struct ntfs_boot_sector*)&buffer,partition,debug,dump_ind)) return 1;
  if(!recover_netware((struct disk_netware *)&buffer,partition)) return 1;
  return 0;
}

int search_type_1(t_param_disk disk_car,t_diskext partition,const int debug, const int dump_ind,const int paranoid)
{
  uchar buffer[8*0x200];
  loff_t hd_offset=partition->lba;
//  ecrit_rapport("search_type_1 lba=%ld\n",partition->lba);
  if(hd_read_LBA(disk_car,8, &buffer, hd_offset))
    return -1;
  if(!recover_BSD(disk_car,(struct disklabel *)&buffer[0x200],partition,debug,dump_ind)) return 1;
  if(!recover_BeFS(disk_car,(struct disk_super_block *)&buffer[0x200],partition,debug,dump_ind)) return 1;
  return 0;
}

int search_type_2(t_param_disk disk_car,t_diskext partition,const int debug, const int dump_ind,const int paranoid)
{
  uchar buffer[8*0x200];
  int i;
  dword ref_lba=partition->lba;
//  ecrit_rapport("search_type_2 lba=%ld\n",partition->lba);
  for(i=0;i<=(paranoid?4:0);i++)
  {
    // If i!=0 (paranoid), try backup superblock
    // It must be in paranoid mode because it can hide otherwise other partition type
    loff_t hd_offset=ref_lba+(i*2*8192);
    if(hd_offset+8<disk_car->lba)
    {
      if(hd_read_LBA(disk_car,8, &buffer, hd_offset))
	return -1;
      partition->lba=hd_offset;
      if(!recover_EXT2(disk_car,(struct ext2_super_block*)&buffer[0x400],partition,debug,dump_ind)) return 1;
      partition->lba=ref_lba;
    }
  }
  return 0;
}

int search_type_128(t_param_disk disk_car,t_diskext partition,const int debug, const int dump_ind,const int paranoid)
{
  if(partition->lba+128<disk_car->lba)
  {
	uchar buffer[8*0x200];
	loff_t hd_offset=partition->lba;
//	ecrit_rapport("search_type_128 lba=%ld\n",partition->lba);
	/* Test ReiserFS */
	if(hd_read_LBA(disk_car,8, &buffer, hd_offset+128)) /* 64k offset */
	  return -1;
	if(!recover_rfs(disk_car,(struct reiserfs_super_block*)&buffer,partition,debug,dump_ind)) return 1;
  }
  return 0;
}

int check_part(t_param_disk disk_car,int debug,t_diskext partition)  
{
  int ret=0;
  switch(partition->part_type)
  {
    case P_12FAT:
    case P_12FATH:
    case P_16FAT:
    case P_16FATH:
    case P_16FATBD:
    case P_16FATBDH:
    case P_32FAT_1:
    case P_32FAT_2:
    case P_32FAT_1H:
    case P_32FAT_2H:
      ret=check_FAT(disk_car,partition,debug);
      break;
    case P_LINUX:
      ret=check_EXT2(disk_car,partition,debug);
      if(ret)
	ret=check_rfs(disk_car,partition,debug);
      if(ret)
      { aff_buffer(BUFFER_ADD,"No EXT2 or Reiser marker\n"); }
      break;
    case P_NTFSH:
    case P_NTFS:
      ret=check_NTFS(disk_car,partition,debug,0);
      break;
    case P_LVM:
      ret=check_LVM(disk_car,partition,debug);
      break;
    case P_BEOS:
      ret=check_BeFS(disk_car,partition,debug);
      break;
    case P_BSD:
      ret=check_BSD(disk_car,partition,debug);
      break;
    case P_LINSWAP:
      ret=check_Linux_SWAP(disk_car,partition,debug);
      break;
  }
  if(ret)
  {
    ecrit_rapport("check_part %02X\n", partition->part_type);
    aff_part_buffer(2,disk_car,partition);
  }
  return ret;
}

