#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>

#ifdef HAVE_E2FS
#include "ext2fs/ext2_fs.h"
#include "ext2fs/ext2fs.h"
#include "types.h"
#include "common.h"
#include "intrface.h"
#include "hdaccess.h"
ext2_filsys     current_fs=NULL;

#define DIRENT_DELETED_FILE	4
/*
 * list directory
 */

#define LONG_OPT	0x0001
#define DELETED_OPT	0x0002

struct list_dir_struct {
	WINDOW *f;
	int	col;
	int	options;
};

static const char *monstr[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
				"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

/*
 * I/O Manager routine prototypes
 */
static errcode_t my_open(const char *dev, int flags, io_channel *channel);
static errcode_t my_close(io_channel channel);
static errcode_t my_set_blksize(io_channel channel, int blksize);
static errcode_t my_read_blk(io_channel channel, unsigned long block, int count, void *buf);
static errcode_t my_write_blk(io_channel channel, unsigned long block, int count, const void *buf);
static errcode_t my_flush(io_channel channel);
static errcode_t my_write_byte(io_channel channel, unsigned long offset, int size, const void *buf);


static io_channel alloc_io_channel(t_param_disk disk_car,t_diskext partition);

static struct struct_io_manager my_struct_manager = {
        EXT2_ET_MAGIC_IO_MANAGER,
        "TestDisk I/O Manager",
        my_open,
        my_close,
        my_set_blksize,
        my_read_blk,
        my_write_blk,
        my_flush,
	my_write_byte
};

io_manager my_io_manager = &my_struct_manager;
/*
 * Macro taken from unix_io.c
 */
/*
 * For checking structure magic numbers...
 */

#define EXT2_CHECK_MAGIC(struct, code) \
          if ((struct)->magic != (code)) return (code)

typedef struct struct_my_data t_my_data;
struct struct_my_data
{
  t_param_disk disk_car;
  t_diskext partition;
};
/*
 * Allocate libext2fs structures associated with I/O manager
 */
static io_channel alloc_io_channel(t_param_disk disk_car,t_diskext partition)
{
  io_channel     ioch;
  t_my_data	*my_data;
#ifdef DEBUG
  ecrit_rapport("alloc_io_channel start\n");
#endif
  ioch = (io_channel)malloc(sizeof(struct struct_io_channel));
  if (!ioch)
	  return NULL;
  memset(ioch, 0, sizeof(struct struct_io_channel));
  ioch->magic = EXT2_ET_MAGIC_IO_CHANNEL;
  ioch->manager = my_io_manager;
  ioch->name = (char *)malloc(strlen(partition->name)+1);
  if (!ioch->name) {
	  free(ioch);
	  return NULL;
  }
  strcpy(ioch->name, partition->name);
  my_data=malloc(sizeof(*my_data));
  my_data->partition=partition;
  my_data->disk_car=disk_car;
  ioch->private_data = my_data;
  ioch->block_size = 1024; /* The smallest ext2fs block size */
  ioch->read_error = 0;
  ioch->write_error = 0;
#ifdef DEBUG
  ecrit_rapport("alloc_io_channel end\n");
#endif
  return ioch;
}

static errcode_t my_open(const char *dev, int flags, io_channel *channel)
{
  *channel = (io_channel*)dev;
#ifdef DEBUG
  ecrit_rapport("my_open done\n");
#endif
  return 0;
};

static errcode_t my_close(io_channel channel)
{
  FREE(channel);
#ifdef DEBUG
  ecrit_rapport("my_close done\n");
#endif
  return 0;
}

static errcode_t my_set_blksize(io_channel channel, int blksize)
{
  channel->block_size = blksize;
#ifdef DEBUG
  ecrit_rapport("my_set_blksize done\n");
#endif
  return 0;
}

static errcode_t my_read_blk(io_channel channel, unsigned long block, int count, void *buf)
{
  size_t size;
  t_my_data *my_data=(t_my_data*)channel->private_data;
  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);

  size = (count < 0) ? -count : count * channel->block_size;
#ifdef DEBUG
  ecrit_rapport("my_read_blk start\n");
  ecrit_rapport("size=%u, block=%u name=%s\n",size,block,my_data->partition->name);
#endif
  if(hd_read_LBA(my_data->disk_car,size/512,buf,my_data->partition->lba+(block*channel->block_size/512)))
    return 1;
#ifdef DEBUG
  ecrit_rapport("my_read_blk done\n");
#endif
  return 0;
}

static errcode_t my_write_blk(io_channel channel, unsigned long block, int count, const void *buf)
{
  t_my_data *my_data=(t_my_data*)channel;
  EXT2_CHECK_MAGIC(channel, EXT2_ET_MAGIC_IO_CHANNEL);
//  if(hd_write_LBA(my_data->disk_car,count*channel->block_size/512,buf,my_data->partition->lba+(block*channel->block_size/512)))
    return 1;
//  return 0;
}

static errcode_t my_flush(io_channel channel)
{
  return 0;
}

static errcode_t my_write_byte(io_channel channel, unsigned long offset, int size, const void *buf)
{
  return 1;
}

static int list_dir_proc(struct ext2_dir_entry *dirent,
			 int	offset,
			 int	blocksize,
			 char	*buf,
			 void	*private)
{
	struct ext2_inode	inode;
	ext2_ino_t		ino;
	struct tm		*tm_p;
	time_t			modtime;
	char			name[EXT2_NAME_LEN];
	char			tmp[EXT2_NAME_LEN + 16];
	char			datestr[80];
	char			lbr, rbr;
	int			thislen;
	struct list_dir_struct *ls = (struct list_dir_struct *) private;

	thislen = ((dirent->name_len & 0xFF) < EXT2_NAME_LEN) ?
		(dirent->name_len & 0xFF) : EXT2_NAME_LEN;
	strncpy(name, dirent->name, thislen);
	name[thislen] = '\0';
	ino = dirent->inode;
#ifdef DEBUG
	ecrit_rapport("name=%s\n",name);
#endif
	lbr = rbr = ' ';
	if (ls->options & LONG_OPT) {
		if (ino) {
			if (ext2fs_read_inode(current_fs,ino, &inode))
				return 0;
			modtime = inode.i_mtime;
			tm_p = localtime(&modtime);
			sprintf(datestr, "%2d-%s-%4d %02d:%02d",
				tm_p->tm_mday, monstr[tm_p->tm_mon],
				1900 + tm_p->tm_year, tm_p->tm_hour,
				tm_p->tm_min);
		} else {
			strcpy(datestr, "                 ");
			memset(&inode, 0, sizeof(struct ext2_inode));
		}
//		aff_buffer(BUFFER_ADD, "%c%6u%c %6o  %5d  %5d   ", lbr, ino, rbr,
//			inode.i_mode, inode.i_uid, inode.i_gid);
		aff_buffer(BUFFER_ADD, "%6o  %5d  %5d   ", 
			inode.i_mode, inode.i_uid, inode.i_gid);
		if (LINUX_S_ISDIR(inode.i_mode))
			aff_buffer(BUFFER_ADD, "%5d", inode.i_size);
		else
			aff_buffer(BUFFER_ADD, "%5lld", inode.i_size |
				((__u64)inode.i_size_high << 32));
		aff_buffer(BUFFER_ADD, " %s %s\n", datestr, name);
	} else {
		sprintf(tmp, "%c%u%c (%d) %s   ", lbr, dirent->inode, rbr,
			dirent->rec_len, name);
		thislen = strlen(tmp);

		if (ls->col + thislen > 80) {
			aff_buffer(BUFFER_ADD, "\n");
			ls->col = 0;
		}
		aff_buffer(BUFFER_ADD, "%s", tmp);
		ls->col += thislen;
	}
	return 0;
}

int dir_ext2(WINDOW *window,t_param_disk disk_car,t_diskext partition)
{
  errcode_t       retval;
  ext2_filsys     fs;
  ext2_ino_t      inode;
  struct list_dir_struct ls;
  int flags;
  const char*name="/";
  io_channel ioch;
  aff_buffer(BUFFER_RESET,"Q");
  wmove(window,4,0);
  aff_part(window,1,disk_car,partition);
  aff_part_rapport(1,disk_car,partition);
  ecrit_rapport("\n");
  ioch=alloc_io_channel(disk_car,partition);
  retval = ext2fs_open (ioch, 0, 0, 0, my_io_manager, &fs);
  if(retval)
  {
    aff_buffer(BUFFER_ADD,"Can't open ext2fs\n");
  }
  else
  {
    current_fs=fs;
    ls.f = window;
    ls.col = 0;
    ls.options = LONG_OPT;
    flags = 0; //DIRENT_FLAG_INCLUDE_EMPTY;
    retval = ext2fs_dir_iterate(fs, EXT2_ROOT_INO, flags, 0, list_dir_proc, &ls);
    if(retval)
    {
      aff_buffer(BUFFER_ADD,"ext2fs_dir_iterate failed\n");
    }
    ext2fs_close (fs);
  }
  aff_buffer(BUFFER_DISPLAY,"Q",window);
  return retval;
}
#else
#include "types.h"
#include "common.h"
#include "intrface.h"
#include "hdaccess.h"

int dir_ext2(WINDOW *window,t_param_disk disk_car,t_diskext partition)
{
  aff_buffer(BUFFER_RESET,"Q");
  move(5,0);
  aff_part(window,0,disk_car,partition);
  aff_part_rapport(0,disk_car,partition);
  aff_buffer(BUFFER_ADD,"Recompile with E2FS library");
  aff_buffer(BUFFER_DISPLAY,"Q",window);
  return 0;
}
#endif
