/*

    File: dir_fat.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <stdio.h>
#include <ctype.h>	/* toupper */
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include "types.h"
#include "common.h"
#include "fat.h"
#include "lang.h"
#include "hdaccess.h"
#include "fnctdsk.h"
#include "testdisk.h"
#include "intrface.h"
int date_dos2unix(unsigned short f_time,unsigned short f_date);
int dir_fat_aux(char*buffer,int size);
static inline void fat16_towchar(wchar_t *dst, const __u8 *src, size_t len);

static inline void fat16_towchar(wchar_t *dst, const __u8 *src, size_t len)
{
#ifdef __BIG_ENDIAN
	while (len--) {
		*dst++ = src[0] | (src[1] << 8);
		src += 2;
	}
#else
	memcpy(dst, src, len * 2);
#endif
}

int dir_fat_aux(char*buffer,int size)
{
  static const char *monstr[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
				"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

  struct msdos_dir_entry *de=(struct msdos_dir_entry*)buffer;
  wchar_t unicode[1000];
  unsigned char long_slots;
GetNew:
  long_slots = 0;
  unicode[0]=0;
  if (de->name[0] == (__s8) DELETED_FLAG)
    goto RecEnd;
  if (de->attr == ATTR_EXT) {
    unsigned int i;
    struct msdos_dir_slot *ds;
    unsigned char id;
    unsigned char slot;
    unsigned char slots;
    unsigned char sum;
    unsigned char alias_checksum;
ParseLong:
    slots = 0;
    ds = (struct msdos_dir_slot *) de;
    id = ds->id;
    if (!(id & 0x40))
      goto RecEnd;
    slots = id & ~0x40;
    if (slots > 20 || !slots)	/* ceil(256 * 2 / 26) */
      goto RecEnd;
    long_slots = slots;
    alias_checksum = ds->alias_checksum;

    slot = slots;
    while (1) {
      int offset;

      slot--;
      offset = slot * 13;
      fat16_towchar(unicode + offset, ds->name0_4, 5);
      fat16_towchar(unicode + offset + 5, ds->name5_10, 6);
      fat16_towchar(unicode + offset + 11, ds->name11_12, 2);

      if (ds->id & 0x40) {
	unicode[offset + 13] = 0;
      }
      de++;
      if(de>=buffer+size)
	goto EODir;
      if (slot == 0)
	break;
      ds = (struct msdos_dir_slot *) de;
      if (ds->attr !=  ATTR_EXT)
	goto RecEnd;	/* XXX */
      if ((ds->id & ~0x40) != slot)
	goto ParseLong;
      if (ds->alias_checksum != alias_checksum)
	goto ParseLong;
    }
    if (de->name[0] == (__s8) DELETED_FLAG)
      goto RecEnd;
    if (de->attr ==  ATTR_EXT)
      goto ParseLong;
    if (IS_FREE(de->name) || (de->attr & ATTR_VOLUME))
      goto RecEnd;
    for (sum = 0, i = 0; i < 11; i++)
      sum = (((sum&1)<<7)|((sum&0xfe)>>1)) + de->name[i];
    if (sum != alias_checksum)
      long_slots = 0;
  }
RecEnd:
  if(!unicode[0])
  {
    int i;
    int j=0;
    for(i=0;(i<8)&&(de->name[i]!=' ');i++)
      unicode[j++]=de->name[i];
    if(de->ext[0]!=' ')
    {
      unicode[j++]='.';
      for(i=0;(i<3)&&(de->ext[i]!=' ');i++)
	unicode[j++]=de->ext[i];
    }
    unicode[j]=0;
  }
  if (unicode[0] && (unicode[0] != (__s8) DELETED_FLAG))
  {
    int i;
    time_t modtime;
    char datestr[80];
    struct tm *tm_p;

    modtime=date_dos2unix(de->time,de->date);
    tm_p = gmtime(&modtime);
    sprintf(datestr, "%2d-%s-%4d %02d:%02d",
      tm_p->tm_mday, monstr[tm_p->tm_mon],
      1900 + tm_p->tm_year, tm_p->tm_hour,
      tm_p->tm_min);

    tm_p = gmtime(&modtime);
    sprintf(datestr, "%2d-%s-%4d %02d:%02d",
      tm_p->tm_mday, monstr[tm_p->tm_mon],
      1900 + tm_p->tm_year, tm_p->tm_hour,
      tm_p->tm_min);
    if (de->attr & ATTR_DIR)
      aff_buffer(BUFFER_ADD,"     <DIR>",de->size);
    else
      aff_buffer(BUFFER_ADD,"%10d",de->size);
    aff_buffer(BUFFER_ADD," %s ",datestr);
    for(i=0;unicode[i];i++)
      aff_buffer(BUFFER_ADD, "%c", unicode[i]);
    aff_buffer(BUFFER_ADD, "\n");
    for(i=0;unicode[i];i++)
      ecrit_rapport("%c", (char)unicode[i]);
    ecrit_rapport("\n");

  }
  de++;
  if(de<buffer+size)
    goto GetNew;
EODir:
  return 0;
}

static int day_n[] = { 0,31,59,90,120,151,181,212,243,273,304,334,0,0,0,0 };
		  /* JanFebMarApr May Jun Jul Aug Sep Oct Nov Dec */

/* Convert a MS-DOS time/date pair to a UNIX date (seconds since 1 1 70). */

int date_dos2unix(unsigned short f_time,unsigned short f_date)
{
	int month,year,secs;

	/* first subtract and mask after that... Otherwise, if
	   f_date == 0, bad things happen */
	month = ((f_date >> 5) - 1) & 15;
	year = f_date >> 9;
	secs = (f_time & 31)*2+60*((f_time >> 5) & 63)+(f_time >> 11)*3600+86400*
	    ((f_date & 31)-1+day_n[month]+(year/4)+year*365-((year & 3) == 0 &&
	    month < 2 ? 1 : 0)+3653);
			/* days since 1.1.70 plus 80's leap day */
	return secs;
}

int dump_fat_root(WINDOW *window,t_param_disk disk_car, t_diskext partition)
{
  t_sector buffer;
  int root_size=0;
  loff_t start=0;
  struct fat_boot_sector*fat_header=(struct fat_boot_sector*)buffer;
  if(hd_read_LBA(disk_car,1, &buffer, partition->lba))
  {
    ecrit_rapport(m_CHKFAT_RERR); return 1;
  }
  if(is_fat32(partition->part_type))
  {
    root_size=fat_header->cluster_size;
    start=partition->lba+fat_header->reserved+fat_header->fats*fat_header->fat32_length+(fat_header->root_cluster-2)*fat_header->cluster_size;
  }
  else
  {
    root_size=((get_dir_entries(fat_header)-1)/16+1);
    start=partition->lba+fat_header->reserved+fat_header->fats*fat_header->fat_length;
  }
  {
    unsigned char buffer_dir[0x200*root_size];
    if(hd_read_LBA(disk_car, root_size, &buffer_dir, start))
    {
      ecrit_rapport(m_ROOT_CLUSTER_RERR); return 1;
    }

//    dump(window,&buffer_dir,0x200*root_size);

    aff_buffer(BUFFER_RESET,"Q");
    wmove(window,4,0);
    aff_part(window,1,disk_car,partition);
    aff_part_rapport(1,disk_car,partition);
    ecrit_rapport("\n");
    dir_fat_aux(buffer_dir,0x200*root_size);

    aff_buffer(BUFFER_DISPLAY,"Q",window);
  }
  return 0;
}
