/*

    File: testdisk.c

    Copyright (C) 1998-2002  Christophe GRENIER <grenier@cgsecurity.org>
  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#include <stdarg.h>
#include <unistd.h>	/* geteuid */
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>
#include <ctype.h>      /* toupper, tolower */
#include "types.h"
#include "common.h"
#include "testdisk.h"
#include "lang.h"
#include "intrface.h"
#include "godmode.h"
#include "fnctdsk.h"

FILE *f_rapport=NULL;

void ecrit_rapport_string(const char *string,const int max_length)
{
  int i;
  for(i=0;(string[i])&&(i<max_length);i++)
    ecrit_rapport("%c",string[i]);
}

int ecrit_rapport(const char *_format, ...)
{
  int res=0;
  if(f_rapport)
  {
    res=vfprintf(f_rapport,_format,(va_list)(&_format+1));
    fflush(f_rapport);
  }
  return res;
}

int doprintf(const char *_format, ...)
{
  char res[800];
#ifdef DEBUG
  if(f_rapport)
    vfprintf(f_rapport,_format,(&_format+1));
#endif
#ifdef DJGPP
  vsprintf(res,_format,(&_format+1));
#else
  vsnprintf(res,sizeof(res),_format,(va_list)(&_format+1));
#endif
  return addstr(res);
}

int wdoprintf(WINDOW *window, const char *_format, ...)
{
  char res[800];
#ifdef DEBUG
  if(f_rapport)
    vfprintf(f_rapport,_format,(&_format+1));
#endif
#ifdef DJGPP
  vsprintf(res,_format,(&_format+1));
#else
  vsnprintf(res,sizeof(res),_format,(va_list)(&_format+1));
#endif
  return waddstr(window,res);
}

int ask_YN(WINDOW *window,const char* msg)
{
  char res;
  wdoprintf(window,msg);
  wrefresh(window);
  do
  {
    res=toupper(getch());
  } while((res!=c_NO)&(res!=c_YES));
  wdoprintf(window,"%c\n",res);
  return (res==c_YES);
}

FILE* init_log(const char*filename,int argc, char**argv)
{
  FILE*f_file=fopen(filename,"a");
  if(f_file==NULL)
	printf(m_LOG_ERR);
  else
  {
	int i;
	time_t my_time=time(NULL);
	fprintf(f_file,"\n\n%s",ctime(&my_time));
	fprintf(f_file,m_CMDLINE);
	for(i=1;i<argc;i++)
	  fprintf(f_file," %s", argv[i]);
	fprintf(f_file,"\n");
  }
  return f_file;
}

int main( int argc, char **argv )
{
  int i;
  unsigned int x;
  int help=0, create_log=0,paranoid=0,debug=0, dump_ind=0;
  int fast_mode=0;
  int align=1;
  int do_analyse=1;
  int do_list=0;
  int test_recovery=0;
  for(i=1;i<argc;i++)
  {
    for (x=0; x < strlen( argv[i] ); x++)
     argv[i][x] = tolower( argv[i][x] );
    /* Deprecated options */
    if (!strcmp(argv[i],"/trust")) ;
    else if (!strcmp(argv[i],"/liar")) ;
    else if (!strcmp(argv[i],"/trust_auto")) ;
    else if (!strcmp(argv[i],"/search"));
    else if(!strcmp(argv[i],"/mbr"));
    else if(!strcmp(argv[i],"/clean"));
    else if(!strcmp(argv[i],"/geometry"));
    /* Some stupid options */
    else if(!strcmp(argv[i],"/nodebug"))
      debug=0;
    else if(!strcmp(argv[i],"/noanalyze"))
      do_analyse=0;
    else if(!strcmp(argv[i],"/noanalyse"))
      do_analyse=0;
    else if(!strcmp(argv[i],"/test_recovery"))
    {
      test_recovery=1;
      do_list=1;
      create_log=1;
    }
    /* */
    else if (!strcmp(argv[i],"/fast"))
      fast_mode=0;
    else if (!strcmp(argv[i],"/medium"))
      fast_mode=1;
    else if (!strcmp(argv[i],"/slow"))
      fast_mode=2;
    else if(!strcmp(argv[i],"/paranoid"))
      paranoid=1;
    else if(!strcmp(argv[i],"/dump"))
      dump_ind=1;
    /* */
    else if(!strcmp(argv[i],"/log"))
      create_log=1;
    else if(!strcmp(argv[i],"/debug"))
    {
      debug=1;
      create_log=1;
    }
    else if(!strcmp(argv[i],"/list"))
      do_list=1;
    else
      help=1;
  }
  if(help)
  {
    printf(m_Copyright);
    printf(m_Usage);
    return 0;
  }
#ifndef DJGPP
  if(geteuid()!=0)
  {
	printf("You need to be root to use TestDisk\n");
	return 1;
  }
#endif
  if(create_log)
  {
    f_rapport=init_log("testdisk.log",argc,argv);
    ecrit_rapport(m_Copyright);
#ifdef DJGPP
    ecrit_rapport("Dos version\n\n");
#elif defined(BSD)
    ecrit_rapport("BSD version\n\n");
#elif defined(LINUX)
    ecrit_rapport("Linux version\n\n");
#else
    ecrit_rapport("Undefined OS\n\n");
#endif
}
printf("Please wait...\n");
if(do_list)
{
  t_param_disk param_disk[NBR_DISK_MAX];
  int nbr_disk;
  for(i=0;i<NBR_DISK_MAX;i++)
  {
	param_disk[i]=(t_param_disk)MALLOC(sizeof(*param_disk[i]));
#ifndef DJGPP
	param_disk[i]->name=NULL;
#endif
  }
  nbr_disk=hd_parse(param_disk,debug,0);
  /* save disk parameters to rapport */
  printf("Hard disk list\n");
  ecrit_rapport("Hard disk list\n");
  for(i=0;i<nbr_disk;i++)
  {
    aff_disk_classic(param_disk[i]);
    aff_disk_rapport(param_disk[i]);
  }
  printf("\n");
  ecrit_rapport("\n");
  for(i=0;i<nbr_disk;i++)
  {
    interface_list(param_disk[i],debug,test_recovery);
    if(test_recovery)
    {
      t_list_part list_part;
      t_list_part element;
      //dump_ind=0;
      list_part=search_part(param_disk[i],paranoid,debug,0,fast_mode,0);
      if(debug)
      {
	ecrit_rapport("search_part() results\n");
	for(element=list_part;element;element=element->next)
	  aff_part_rapport(1,param_disk[i],element->part);
      }
      delete_list_part(list_part);
    }
  }
  for(i=0;i<NBR_DISK_MAX;i++)
  {
#ifndef DJGPP
	if(param_disk[i]->name)    /* forget name */
	  FREE(param_disk[i]->name);
#endif
	FREE(param_disk[i]);
  }
} else
  do_curses_testdisk(debug,paranoid,dump_ind,fast_mode,align,do_analyse);
#ifndef DJGPP
  printf("\n");
#endif
  if(f_rapport)
	fclose(f_rapport);
  return 0;
}
